#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "pico.h"

extern int lexval, nsrc;
int lat; /* look ahead token */
extern char text[];
extern void emit(int);
extern Biobuf *bstdin;

void expr(void);

void expect(int t)
{
	if (lat == t)
		lat = lex();
	else
		error("error: expected token %T\n", t);
}

void fileref(int val, int tok)
{
	if (val < 0 || val > nsrc)
		error("bad file number: %d\n", val);
	emit(VALUE);
	emit(val);
	if (lat == '[') {
		lat = lex();
		expr();
		expect(',');
		expr();
		expect(']'); /* [x,y] */
	} else {
		emit('x');
		emit('y');
	}
	if (lat == '.') {
		lat = lex();
		if (lat != SPECIAL_RED && lat != SPECIAL_GREEN && lat != SPECIAL_BLUE)
			error("expected red, green, blue, bad token %T\n", lat);
		if (tok == RVAL)
			emit(lat);
		else
			switch (lat) {
			case SPECIAL_RED:
				emit(SPECIAL_LRED);
				break;
			case SPECIAL_GREEN:
				emit(SPECIAL_LGREEN);
				break;
			case SPECIAL_BLUE:
				emit(SPECIAL_LBLUE);
				break;
			}
		lat = lex();
	} else
		emit(tok);
}

void factor(void)
{
	int n;
	char *s;
	Symbol *sym;

	switch (lat) {
	case '(':
		lat = lex();
		expr();
		expect(')');
		break;
	case '[':
		lat = lex();
		expr();
		expect(',');
		expr();
		expect(',');
		expr();
		expect(']');
		emit(MKCOLOR);
		break;
	case '-':
		lat = lex();
		expr();
		emit(UMIN);
		break;
	case '!':
		lat = lex();
		factor();
		emit('!');
		break;
	case OLD:
		lat = lex();
		fileref(CUROLD, RVAL);
		break;
	case NAME:
	case ARRAY:
		s = strdup(text);
		lat = lex();
		if (lat == '[') { /* array */
			lat = lex();
			expr();
			expect(']');
			if ((sym = lookup(s)) == nil) {
				sym = install(s, ARRAY, 0);
				sym->arr.len = 2;
				sym->arr.mem = emalloc(sym->arr.len * sizeof (long));
			} else if (sym->type != ARRAY) {
				sym->type = ARRAY;
				sym->arr.len = 2;
				sym->arr.mem = emalloc(sym->arr.len * sizeof (long));
			}
			emit(VALUEOFAT);
		} else {
			if ((sym = lookup(s)) == nil)
				sym = install(s, NAME, 0);
			emit(VALUEOF);
		}
		emit((int) sym);
		break;
	case '$':
		lat = lex();
		if (lat == VALUE) {
			expect(VALUE);
			fileref(lexval + 1, RVAL);
			break;
		}
		/* else try $name by ignoring $ and going right to the... */
	case FNAME:
		n = lexval;
		lat = lex();
		fileref(n + 1, RVAL);
		break;
	case '~':
		lat = lex();
		expr();
		emit('~');
		break;
	case SPECIAL_X:
	case SPECIAL_Y:
	case SPECIAL_Z:
		emit(lat);
		lat = lex();
		break;
	case VALUE:
		emit(VALUE);
		emit(lexval);
		lat = lex();
		break;
	case SPECIAL_y:
	case SPECIAL_x:
		emit((lat == SPECIAL_y) ? 'y' : 'x');
		lat = lex();
		break;
	default:
		error("expr: syntax error, token %T\n", lat);
	}
	if (lat == POW) {
		lat = lex();
		factor();
		emit(POW);
	}
}

void multiplicative(void)
{
	int save;

	factor();
	while ((lat == '*') || (lat == '/') || (lat == '%')) {
		save = lat;
		lat = lex();
		factor();
		emit(save);
	}
}

void additive(void)
{
	int save;

	multiplicative();
	while ((lat == '+') || (lat == '-')) {
		save = lat;
		lat = lex();
		multiplicative();
		emit(save);
	}
}

void shift(void)
{
	int save;

	additive();
	while ((lat == LSH) || (lat == RSH)) {
		save = lat;
		lat = lex();
		additive();
		emit(save);
	}
}

void relational(void)
{
	int save;

	shift();
	while ((lat == '<') || (lat == '>') || (lat == GE) || (lat == NE)) {
		save = lat;
		lat = lex();
		shift();
		emit(save);
	}
}

void equality(void)
{
	int save;

	relational();
	while ((lat == EQ) || (lat == NE)) {
		save = lat;
		lat = lex();
		relational();
		emit(save);
	}
}

void bitwiseand(void)
{
	equality();
	while (lat == '&') {
		lat = lex();
		equality();
		emit('&');
	}
}

void bitwisexor(void)
{
	bitwiseand();
	while (lat == '^') {
		lat = lex();
		bitwiseand();
		emit('^');
	}
}

void bitwiseor(void)
{
	bitwisexor();
	while (lat == '|') {
		lat = lex();
		bitwisexor();
		emit('|');
	}
}

void logical(void)
{
	int save;

	bitwiseor();
	while ((lat == AND) || (lat == OR)) {
		save = lat;
		lat = lex();
		bitwiseor();
		emit(save);
	}
}

void conditional(void)
{
	extern int prs;
	extern int parsed[MANY];
	int remem1, remem2;

	logical();
	if (lat == '?') {
		lat = lex();
		emit('?');
		remem1 = prs;
		emit(0);
		expr();
		expect(':');
		emit(':');
		remem2 = prs;
		emit(0);
		parsed[remem1] = prs - 1;
		expr();
		parsed[remem2] = prs - 1;
	}
}

void expr(void)
{
	char *n;
	Symbol *s;

	if (lat == NAME) {
		n = strdup(text);
		lat = lex();
		if (lat == '=') {
			lat = lex();
			expr();
			if ((s = lookup(n)) == nil)
				s = install(n, NAME, 0);
			emit(ASSIGN);
			emit((int) s);
			return;
		} else if (lat == '[') {
			saveloc();
			lat = lex();
			expr();
			expect(']');
			if ((s = lookup(n)) == nil) {
				s = install(n, ARRAY, 0);
				s->arr.len = 2;
				s->arr.mem = emalloc(s->arr.len * sizeof (long));
			} else if (s->type != ARRAY) {
				s->type = ARRAY;
				s->arr.len = 2;
				s->arr.mem = emalloc(s->arr.len * sizeof (long));
			}
			if (lat == '=') {
				lat = lex();
				expr();
				emit(ASSIGNARR);
				emit((int) s);
				return;
			} else {
				putbackto();
				putback(strlen(n));
			}
			lat = lex();
		} else {
			putback(1);
			putback(strlen(n));
			lat = lex();
		}
	}
	conditional();
}

void transform(void)
{
	extern int prs;

	prs = 0; /* initial length of parse string */
	if (lat != NEW) {
		expr();
		emit('@');
		putback(1);
		return;
	}
	lat = lex();
	if (lat == '[') {
		fileref(CURNEW, LVAL);
		expect('=');
		expr();
		emit('=');
	} else {
		expect('=');
		expr();
		emit('@');
	}
	if (lat != '\n' && lat != ';')
		error("syntax error, separator\n");
	putback(1);
}
