#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>
#include <ctype.h>
#include "pico.h"

extern int nsrc, cropx, cropy;

static char *basename(char *s)
{
	char *pr, *t;

	pr = strrchr(s, '/');
	if (pr == nil)
		pr = s;
	else
		pr++;
	t = strdup(pr);
	if (t == nil)
		error("memory exhausted\n");
	pr = t;
	if (isdigit(*pr)) {
		char *u;

		u = smprint("_%s", t);
		if (u == nil)
			error("memory exhausted\n");
		free(t);
		pr = t = u;
	}
	while (isalnum(*pr) || (*pr) == '_')
		pr++;
	*pr = '\0';
	return t;
}

static void _up(struct SRC *what, struct SRC *to)
{
	int doup = 0, oldh, oldenc, i;

	SET(oldh);
	SET(oldenc);
	if (what->h != to->h) {
		oldh = what->h;
		what->h = to->h;
		doup = 1;
	}
	if (what->w != to->w) {
		what->w = to->w;
		doup = 1;
	}
	if (to->encoding != what->encoding) {
		oldenc = what->encoding;
		what->encoding = to->encoding;
		doup = 1;
	}
	if (doup) {
		if (oldenc == RGBA32) {
			for (i = 0; i < oldh; i++)
				free(what->pix[i]);
			free(what->pix);
		} else {
			for (i = 0; i < oldh; i++)
				free(what->bpix[i]);
			free(what->bpix);
		}
		if (what->encoding == RGBA32) {
			what->pix = (u32int **) emalloc(what->h * sizeof (u32int *));
			for (i = 0; i < what->h; i++)
				what->pix[i] = (u32int *) emalloc(what->w * sizeof (u32int));
		} else {
			what->bpix = (uchar **) emalloc(what->h * sizeof (uchar *));
			for (i = 0; i < what->h; i++)
				what->bpix[i] = (uchar *) emalloc(what->w * sizeof (uchar));
		}
		what->cx = cropx;
		what->cy = cropy;
	}
}

static void updateoldnew(struct SRC *to)
{
	_up(&src[CUROLD], to);
	_up(&src[CURNEW], to);
}

void getpix(struct SRC *into, char *str)
{
	int fd, i, j;
	Memimage *m;
	extern int docrop, cropw, croph;

	if ((fd = open(str, OREAD)) < 0)
		error("open: %r\n");
	m = readmemimage(fd);
	if (m == nil)
		error("could not read image: %r\n");
	if (into->encoding == RGBA32) {
		if (into->pix != (u32int **) 0) {
			int i;

			for (i = 0; i < into->h; i++)
				free(into->pix[i]);
			free(into->pix);
		}
	} else {
		if (into->bpix != (uchar **) 0) {
			int i;

			for (i = 0; i < into->h; i++)
				free(into->bpix[i]);
			free(into->bpix);
		}
	}
	into->w = docrop ? cropw : m->r.max.x;
	into->h = docrop ? croph : m->r.max.y;
	into->cx = cropx;
	into->cy = cropy;
	into->str = basename(str);
	if ((m->chan != RGBA32) && (m->chan != GREY8)) { /* not proper format - convert */
		Memimage *n;

		n = allocmemimage(m->r, RGBA32); /* blindly assume RGBA32 */
		if (n == nil)
			error("memory exhausted\n");
		memimagedraw(n, n->r, m, m->r.min, nil, ZP, S);
		freememimage(m);
		m = n;
	}
	into->encoding = m->chan;
	if (into->encoding == RGBA32) {
		into->pix = (u32int **) emalloc(into->h * sizeof (u32int *));
		for (i = 0; i < into->h; i++)
			into->pix[i] = (u32int *) emalloc(into->w * sizeof (u32int));
	} else {
		into->bpix = (uchar **) emalloc(into->h * sizeof (uchar *));
		for (i = 0; i < into->h; i++)
			into->bpix[i]= (uchar *) emalloc(into->w * sizeof (uchar));
	}
	updateoldnew(into);
	for (i = cropy; i < into->h; i++)
		for (j = cropx; j < into->w; j++)
			if (into->encoding == RGBA32) {
				ulong *p;

				p = wordaddr(m, Pt(j, i));
				if (p == nil)
					error("could not read image: %r\n");
				into->pix[i - cropy][j - cropx] = (u32int) (*p);
			} else {
				uchar *p;

				p = byteaddr(m, Pt(j, i));
				if (p == nil)
					error("could not read image: %r\n");
				into->bpix[i - cropy][j - cropx] = *p;
			}
	close(fd);
	nsrc++;
}

void putpix(struct SRC *into, char *str)
{
	int fd, i, j;
	Memimage *m;

	if ((fd = create(str, OWRITE, 0777)) < 0)
		error("create: %r\n");
	m = allocmemimage(Rect(0, 0, into->w - into->cx, into->h - into->cy), into->encoding);
	if (m == nil)
		error("memory exhausted\n");
	for (i = 0; i < into->h - into->cy; i++)
		for (j = 0; j < into->w - into->cx; j++)
			if (into->encoding == RGBA32) {
				ulong *p;

				p = wordaddr(m, Pt(j, i));
				if (p == nil)
					error("could not write image: %r\n");
				*p = ((((ulong) (into->pix[i][j])) >> 8) << 8) + 0xFF;
			} else {
				uchar *p;

				p = byteaddr(m, Pt(j, i));
				if (p == nil)
					error("could not write image: %r\n");
				*p = (into->bpix[i][j]);
			}
	if (writememimage(fd, m) < 0)
		error("could not write image: %r\n");
	close(fd);
}

void showfiles(void)
{
	int n;

	if (nsrc == 2)
		print("no files open\n");
	else
		for (n = 2; n < nsrc; n++)
			print("$%d\t%s\t%s\n", n-1, src[n].str,
				(src[n].encoding == RGBA32) ? "color" : "black/white");
}
