#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "pico.h"

extern int nsrc, lexval;
extern char text[];
extern Biobuf *bstdin;

static char *line = nil;
static char *lineptr;
static char *lineptrsv = nil;
int symbolize = 1;

void saveloc(void)
{
	lineptrsv = lineptr;
}

void putbackto(void)
{
	if (lineptrsv == nil)
		return;
	lineptr = lineptrsv;
	lineptrsv = nil;
}

void putback(int count)
{
	if (count == 1)
		lineptr--;
	else {
		while (*lineptr == ' ' || *lineptr == '\t')
			lineptr--;
		while (count-- >= -1)
			lineptr--;
	}
}

int getnumber(int first)
{
	int c;

	lexval = first - '0';
	while (isdigit(c = *lineptr++))
		lexval = 10 * lexval + c - '0';
	lineptr--;;
	return VALUE;
}

int getstring(int first)
{
	int c = first;
	char *str = text;
	Symbol *s;

	if (first == '\"') {
		while ((c = *lineptr++) != '\"')
			*str++ = c;
	} else {
		do {
			*str++ = c;
			c = *lineptr++;
		} while (isalpha(c) || c == '_' || isdigit(c));
	}
	*str = '\0';
	if (first != '\"')
		lineptr--;
	for (c = 2; c < nsrc; c++)
		if (!strcmp(src[c].str, text)) {
			lexval = c - 1;
			return FNAME;
		}
	if (first == '\"')
		return STRING;
	if ((s = lookup(text)) == nil)
		if (symbolize)
			s = install(text, NAME, 0);
	return symbolize ? s->type : ((s != nil) ? s->type : first);
}

int follow(int tok, int ifyes, int ifno)
{
	int c;

	if ((c = *lineptr++) == tok) {
		USED(c);
		return ifyes;
	}
	lineptr--;
	return ifno;
}

int lex(void)
{
	int c;

	if (line != nil)
		if (*lineptr == '\0') {
			free(line);
			line = nil;
		}
	if (line == nil) {
		line = Brdstr(bstdin, '\n', 0);
		if (line == nil) {
			if (bstdin->state != Bractive) /* Woah there - what do you think you are doing, accessing internal members of a Biobuf? Well, I don't know of another way to check for eof, since Brdstr returns nil both on error and on eof. This is taken from the Brdstr source code. */
				error("memory exhausted\n");
			/* otherwise, user hit ctrl+d */
			return 'q';
		}
		lineptr = line;
	}
	do { /* ignore white space */
		c = *lineptr++;
	} while (c == ' ' || c == '\t');
	if (isdigit(c))
		c = getnumber(c);
	else if (isalpha(c) || c == '_' || c == '\"')
		c = getstring(c);
	switch (c) {
	case Beof:
		c = 'q';
		break;
	case '*':
		c = follow('*', POW, c);
		break;
	case '>':
		c = follow('=', GE, follow('>', RSH, c));
		break;
	case '<':
		c = follow('=', LE, follow('<', LSH, c));
		break;
	case '!':
		c = follow('=', NE, c);
		break;
	case '=':
		c = follow('=', EQ, c);
		break;
	case '|':
		c = follow('|', OR, c);
		break;
	case '&':
		c = follow('&', AND, c);
		break;
	}
	return c;
}
