#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <draw.h>
#include <memdraw.h>
#include <bio.h>
#include "pico.h"

int parsed[MANY];
struct SRC src[MANY];
short CUROLD = 0, CURNEW = 1;
int lexval, prs = 0, nsrc = 2;
int docrop = 0, cropx = 0, cropy = 0, cropw = 0, croph = 0;
char text[256];
Biobuf *bstdin;
int lineno = 1;
static jmp_buf again;

int Tfmt(Fmt *f) /* format %T arguments - %T = token */
{
	int i;

	i = va_arg(f->args, int);
	if (isprint(i))
		return fmtprint(f, "%c", (char) i);
	else
		return fmtprint(f, "%d", i);
}

void error(char *s, ...)
{
	extern int lat;
	va_list arg;

	va_start(arg, s);
	vfprint(2, s, arg);
	va_end(arg);
	while (lat != '\n')
		lat = lex();
	lineno++;
	longjmp(again, 1);
}

void *emalloc(ulong N)
{
	void *try;

	if ((try = malloc(N)) == nil)
		error("memory exhausted\n");
	return try;
}

void *erealloc(void *M, ulong N)
{
	void *try;

	if ((try = realloc(M, N)) == nil)
		error("memory exhausted\n");
	return try;
}

void emit(int what)
{
	if (prs >= MANY)
		error("expression too long\n");
	parsed[prs++] = what;
}

void getname(int checkcrop)
{
	int t;
	extern int lexval;

	t = lex();
	if (checkcrop)
		if (t == VALUE) {
			cropx = lexval;
			t = lex();
			if (t == VALUE) {
				cropy = lexval;
				t = lex();
				if (t == VALUE) {
					cropw = lexval;
					t = lex();
					if (t == VALUE) {
						croph = lexval;
						t = lex();
					} else
						error("expected height, got bad token: %T\n", t);
				} else
					error("expected width, got bad token: %T\n", t);
			} else
				error("expected y, got bad token: %T\n", t);
			docrop = 1;
		}
	if (t != NAME && t != FNAME && t != COMMAND_SHOW && t != STRING && !isalpha(t))
		error("expected name, bad token: %T\n", t);
}

int parse(void)
{
	extern int lat; /* look ahead token */

	docrop = cropx = cropy = cropw = croph = once = 0;
	print("%d:\t", lineno);
	for (;;) {
		symbolize = 0;
		lat = lex();
		symbolize = 1;
		if (lat == COMMAND_SHOW) {
			lat = lex();
			if (lat == '\n') {
				showsymbols();
				putback(1); /* put back newline */
			} else {
				Symbol *s;

				if ((s = lookup(text)) == nil)
					print("no such symbol\n");
				else
					print("%s = %ld\n", s->name, s->value);
			}
			continue;
		} else if (lat == 'q') /* eof */
			return 0;
		else if (lat == '\n') {
			lineno++;
			return 1;
		} else if (lat == ';')
			continue;
		else if (strlen(text) == 1)
			switch (*text) {
			case 'q': /* literal q */
				return 0;
			case 'f':
				showfiles();
				break;
			case 'u':
				CUROLD = CURNEW;
				CURNEW = 1 - CUROLD;
				break;
			case 'r':
			case 'a':
				symbolize = 0;
				getname(1);
				symbolize = 1;
				getpix(&src[nsrc], text);
				break;
			case 'w':
				symbolize = 0;
				getname(0);
				symbolize = 1;
				putpix(&src[CUROLD], text);
				break;
			case 'x':
				lat = lex();
				if (lat == '{') {
					once = 1;
					lat = lex();
				}
				transform();
				if (once) {
					extern void expect(int);

					expect('}');
				}
				run();
				break;
			default:
				error("unknown command ID\n");
			}
		else
			error("unknown command ID\n");
	}
}

void main(int argc, char *argv[])
{
	int i;

	fmtinstall('T', Tfmt);
	bstdin = emalloc(sizeof (Biobuf));
	if (Binit(bstdin, 0, OREAD) < 0)
		sysfatal("could not intialize buffer for stdin: %r");
	memimageinit();
	src[CUROLD].pix = (u32int **) emalloc(DEF_Y * sizeof (u32int *));
	src[CURNEW].pix = (u32int **) emalloc(DEF_Y * sizeof (u32int *));
	for (i = 0; i < DEF_Y; i++) {
		src[CUROLD].pix[i] = (u32int *) emalloc(DEF_X * sizeof (u32int));
		src[CURNEW].pix[i] = (u32int *) emalloc(DEF_X * sizeof (u32int));
	}
	src[CUROLD].w = src[CURNEW].w = DEF_X;
	src[CUROLD].h = src[CURNEW].h = DEF_Y;
	src[CUROLD].encoding = src[CURNEW].encoding = RGBA32;
	initsymbols();
	if (setjmp(again) == 0) /* only parse arguments if we called setjmp */
		for (i = 1; i < argc; i++)
			getpix(&src[nsrc], argv[i]);
	else
		fprint(2, "error reading files on command line - rest of line ignored,\nuse f to see what we did get\n");
	do { setjmp(again); } while (parse());
	exits(0);
}
