#include <u.h>
#include <libc.h>
#include <draw.h>
#include "pico.h"

extern int prs, parsed[MANY];
int once = 0;

#define A ((enc == RGBA32) ? ((long) a) : ((uchar) a))
#define B ((enc == RGBA32) ? DBlack : 0)
#define dop(OP) a = *--rr; tr = rr-1; *tr = (*tr OP (long) a)
#define dopd(OP) a = *--rr; tr = rr-1; *tr = a ? (*tr OP (long) a) : B

long Pow(long a, long b)
{
	return (long) pow((double) a, (double) b);
}

void run(void)
{
	long R[MANY];		/* the stack */
	register long *rr, *tr;	/* top of stack */
	register u32int *u;	/* explicit destination color */
	register uchar *uc;	/* explicit destination b/w */
	register u32int *p;	/* default destination color */
	register uchar *pc;	/* default destination b/w */
	register int k;		/* indexes parse string */
	int a, b, c;			/* scratch */
	int x, y;			/* coordinates */
	Symbol *s;		/* symbol */
	RGB color;		/* color */
	int enc;			/* encoding */

	SET(p);
	SET(pc);
	if ((enc = src[CURNEW].encoding) == RGBA32)
		p = src[CURNEW].pix[0];
	else
		pc = src[CURNEW].bpix[0];
	for (y = 0; y < src[CURNEW].h; y++) {
		for (x = 0; x < src[CURNEW].w; x++) {
			for (k = 0, rr = R; k < prs; k++) {
				switch (parsed[k]) {
				case VALUE:
					*rr++ = (ulong) parsed[++k];
					continue;
				case VALUEOF:
					s = (Symbol *) parsed[++k];
					*rr++ = s->value;
					continue;
				case VALUEOFAT:
					s = (Symbol *) parsed[++k];
					a = *--rr;
					if (s->arr.len <= a) {
						s->arr.len = a + 1;
						s->arr.mem = erealloc(s->arr.mem, s->arr.len * sizeof (long));
					}
					*rr++ = s->arr.mem[a];
					continue;
				case ASSIGN: /* assign to variable */
					s = (Symbol *) parsed[++k];
					s->value = *--rr;
					continue;
				case ASSIGNARR: /* assign to array index */
					s = (Symbol *) parsed[++k];
					a = *--rr;
					b = *--rr;
					if (s->arr.len <= b) {
						s->arr.len = b + 1;
						s->arr.mem = erealloc(s->arr.mem, s->arr.len * sizeof (long));
					}
					s->arr.mem[b] = a;
					continue;
				case '@': /* either assigning to general image or normal expression */
					a = *--rr;
					if (enc == RGBA32)
						*p = (u32int) a;
					else
						*pc = (uchar) a;
					continue;
				case SPECIAL_Z:
					*rr++ = ((src[CURNEW].encoding == GREY8) ? 255 : (DWhite));
					continue;
				case SPECIAL_Y:
					*rr++ = (src[CURNEW].h);
					continue;
				case SPECIAL_X:
					*rr++ = (src[CURNEW].w);
					continue;
				case '+':
					dop(+);
					break;
				case '-':
					dop(-);
					break;
				case '*':
					dop(*);
					break;
				case '/':
					dopd(/);
					break;
				case '%':
					dop(%);
					break;
				case '>':
					dop(>);
					break;
				case '<':
					dop(<);
					break;
				case GE:
					dop(>=);
					break;
				case LE:
					dop(<=);
					break;
				case EQ:
					dop(==);
					break;
				case NE:
					dop(!=);
					break;
				case AND:
					dop(&&);
					break;
				case OR:
					dop(||);
					break;
				case '&':
					dop(&);
					break;
				case '|':
					dop(|);
					break;
				case '^':
					dop(^);
					break;
				case LSH:
					dop(<<);
					break;
				case RSH:
					dop(>>);
					break;
				case '~':
					tr = rr - 1;
					*tr = ~(*tr);
					break;
				case 'x':
					*rr++ = (long) x;
					break;
				case 'y':
					*rr++ = (long) y;
					break;
				case UMIN:
					tr = rr - 1;
					*tr = -(*tr);
					break;
				case '!':
					tr = rr - 1;
					*tr = !(*tr);
					break;
				case '=': /* assign to specific pixel */
					a = *--rr;
					if (enc == RGBA32) {
						u = (u32int *) (*--rr);
						*u = (u32int) a;
					} else {
						uc = (uchar *) (*--rr);
						*uc = (uchar) a;
					}
					break;
				case RVAL:
					a = *--rr;
					b = *--rr;
					tr = rr - 1;
					c = *tr;
					if (b >= src[c].w) { /* if x out of bounds, wrap around */
						a++;
						b -= src[c].w;
					} else if (b < 0) /* b negative - make b zero */
						b = 0;
					if ((a < 0) || (a >= src[c].h)) /* if y out of bounds, black, else pixel */
						*tr = DBlack;
					else if (src[c].encoding == GREY8)
						*tr = (src[c].bpix[a][b]);
					else
						*tr = (ulong) (src[c].pix[a][b]);
					enc = src[c].encoding;
					break;
				case SPECIAL_RED:
					a = *--rr;
					b = *--rr;
					tr = rr - 1;
					c = *tr;
					if (src[c].encoding == GREY8)
						error("cannot get color for black/white image\n");
					if (b >= src[c].w) { /* if x out of bounds, wrap around */
						a++;
						b -= src[c].w;
					} else if (b < 0) /* b negative - make b zero */
						b = 0;
					if ((a < 0) || (a >= src[c].h)) /* if y out of bounds, black, else pixel */
						*tr = DRed;
					else
						*tr = (src[c].pix[a][b] & DRed) >> 24;
					break;
				case SPECIAL_GREEN:
					a = *--rr;
					b = *--rr;
					tr = rr - 1;
					c = *tr;
					if (src[c].encoding == GREY8)
						error("cannot get color for black/white image\n");
					if (b >= src[c].w) { /* if x out of bounds, wrap around */
						a++;
						b -= src[c].w;
					} else if (b < 0) /* b negative - make b zero */
						b = 0;
					if ((a < 0) || (a >= src[c].h)) /* if y out of bounds, black, else pixel */
						*tr = DGreen;
					else
						*tr = (src[c].pix[a][b] & DGreen) >> 16;
					break;
				case SPECIAL_BLUE:
					a = *--rr;
					b = *--rr;
					tr = rr - 1;
					c = *tr;
					if (src[c].encoding == GREY8)
						error("cannot get color for black/white image\n");
					if (b >= src[c].w) { /* if x out of bounds, wrap around */
						a++;
						b -= src[c].w;
					} else if (b < 0) /* b negative - make b zero */
						b = 0;
					if ((a < 0) || (a >= src[c].h)) /* if y out of bounds, black, else pixel */
						*tr = DBlue;
					else
						*tr = (src[c].pix[a][b] & DBlue) >> 8;
					break;
				case LVAL:
					a = *--rr;
					b = *--rr;
					tr = rr - 1;
					c = *tr;
					if (b >= src[c].w) { /* if x out of bounds, wrap around */
						a++;
						b -= src[c].w;
					} else if (b < 0) /* b negative - make b zero */
						b = 0;
					if ((a < 0) || (a >= src[c].h)) /* if y out of bounds, error, else pixel */
						error("y out of bounds\n");
					else if (src[c].encoding == GREY8)
						*tr = (long) &(src[c].bpix[a][b]);
					else
						*tr = (long) &(src[c].pix[a][b]);
					enc = src[c].encoding;
					break;
				case SPECIAL_LRED:
				case SPECIAL_LGREEN:
				case SPECIAL_LBLUE:
					error("not implemented\n", 0);
					break;
				case POW:
					a = *--rr;
					*(rr - 1) = Pow(*(rr - 1), a);
					break;
				case '?':
					a = *--rr;
					k++;
					if (!a)
						k = parsed[k];
					break;
				case ':':
					k = parsed[k + 1];
					break;
				case MKCOLOR:
					color.blue = (*--rr);
					color.green = (*--rr);
					tr = rr - 1;
					color.red = (*tr);
					*tr = (color.red & 0xFF) << 24;
					*tr |= (color.green & 0xFF) << 16;
					*tr |= (color.blue & 0xFF) << 8;
					*tr |= 0xFF;
					break;
				default:
					error("run: unknown operator\n");
				}
				if (once)
					goto onceout;
			}
			if (enc == RGBA32)
				p++;
			else
				pc++;
		}
		if (enc == RGBA32)
			p = src[CURNEW].pix[y + 1];
		else
			pc = src[CURNEW].bpix[y + 1];
	}
onceout:
	CUROLD = CURNEW;
	CURNEW = 1 - CUROLD;
}
