#include <u.h>
#include <libc.h>
#include "pico.h"

static Symbol *symtbl;
static Symbol *keytbl;

Symbol *lookup(char *name)
{
	Symbol *s;

	for (s = keytbl; s != nil; s = s->next)	/* 1) look in keywords */
		if (!strcmp(s->name, name))
			return s;
	for (s = symtbl; s != nil; s = s->next)	/* 2) look in symbols */
		if (!strcmp(s->name, name))
			return s;
	return nil; /* not found */
}

Symbol *klookup(char *name)
{
	Symbol *s;

	for (s = keytbl; s != nil; s = s->next)
		if (!strcmp(s->name, name))
			return s;
	return nil; /* not found */
}

static void kinstall(char *name, int type, long value)
{
	Symbol *s;

	s = emalloc(sizeof (Symbol));
	s->name = strdup(name);
	s->type = type;
	s->value = value;
	s->next = keytbl;
	keytbl = s;
}

Symbol *install(char *name, int type, long value)
{
	Symbol *s;

	if (klookup(name) != nil)
		error("cannot redefine keyword\n");
	s = emalloc(sizeof (Symbol));
	s->name = strdup(name);
	s->type = type;
	s->value = value;
	s->next = symtbl;
	symtbl = s;
	return s;
}

void initsymbols(void)
{
	kinstall("new", NEW, 0);
	kinstall("old", OLD, 0);
	kinstall("x", SPECIAL_x, 0);
	kinstall("y", SPECIAL_y, 0);
	kinstall("X", SPECIAL_X, 0);
	kinstall("Y", SPECIAL_Y, 0);
	kinstall("Z", SPECIAL_Z, 0);
	kinstall("red", SPECIAL_RED, 0);
	kinstall("green", SPECIAL_GREEN, 0);
	kinstall("grn", SPECIAL_GREEN, 0);
	kinstall("blue", SPECIAL_BLUE, 0);
	kinstall("blu", SPECIAL_BLUE, 0);
	kinstall("show", COMMAND_SHOW, 0);
}

void showsymbols(void)
{
	Symbol *s;

	for (s = symtbl; s != nil; s = s->next)
		if (s->type == ARRAY) {
			ulong i;

			print("%s = (%uld) ", s->name, s->arr.len);
			for (i = 0; i < s->arr.len; i++)
				print("%ld ", s->arr.mem[i]);
			print("\n");
		} else
			print("%s = %ld\n", s->name, s->value);
}
